//=============================================================================
// INTMATH CLASS
//=============================================================================

#include "intmath.h"

#include "labels.h"
#include "assemble.h"


#undef MATHDBG


//-----------------------------------------------------------------------------
//---= tables =----------------------------------------------------------------
//-----------------------------------------------------------------------------
char *imo[_IMO_LAST+1] =	//all operator characters
{
	//highest presedence (last execution)

	//LOGICAL operators
	"&&", "||",
	"==", "!=",
	">=", "<=",
	">",  "<",

	//MATH operators
	"+",  "-",
	"*",  "/",  "%",

	//BITWISE operators
	"<<", ">>",
	"&",  "!&",
	"|",  "!|",
	"^",  "!^",
	"!",

	"%HI", "%LO",

	//lowest presedence (first execution)

	""
};

__int8 imo_prio[_IMO_LAST+1] =
    //define a low value to execute instruction at last
	//and a higher value for a higher presedence
{
	0,0,
	2,2,
	2,2,
	2,2,

	4,4,
	5,5,5,

	3,3,

	1,1,
	1,1,
	1,1,

	6,

	7,7,

	8 //highest value at the end
};

__int8 imo_type[_IMO_LAST+1] =
{
	IMO_ASSOC_LR,	IMO_ASSOC_LR,
	IMO_ASSOC_LR,	IMO_ASSOC_LR,
	IMO_ASSOC_LR,	IMO_ASSOC_LR,
	IMO_ASSOC_LR,	IMO_ASSOC_LR,

	IMO_ASSOC_LR,	IMO_ASSOC_LDEF,
	IMO_ASSOC_LR,	IMO_ASSOC_LR,

	IMO_ASSOC_LR,	IMO_ASSOC_LR,	IMO_ASSOC_LR,
	IMO_ASSOC_LR,	IMO_ASSOC_LR,
	IMO_ASSOC_LR,	IMO_ASSOC_LR,
	IMO_ASSOC_LR,	IMO_ASSOC_LR,

	IMO_ASSOC_R,
	IMO_ASSOC_R,	IMO_ASSOC_R,

	_IMO_ASSOC_LAST
};


//-----------------------------------------------------------------------------
// INTMATH::is_value: returns true if its a value
//-----------------------------------------------------------------------------
BOOL INTMATH::is_value(char *str)
{
	int		lcc = strlen(str)-1;	//points to last char

	if(lcc < 0) return(false);

	if(str[0] == '$')
	{
		if(!enable_$) return(false);
		str++;
		if(lcc < 1) return(false);
	}
	else if((str[0] == '0')&&(str[1] == 'x'))
	{
		str+=2;
		if((--lcc) < 1) return(false);
	}
	else if(str[0] == '\'')
	{
		if((!lcc)||(str[0]!='\'')) return(false);
		while(*(++str))
		{
			if(*str=='\'')
			{
				switch(*(++str))
				{
				case NULL: return(true);
				case '\'': break;
				default: return(false);
				}
			}
		}
		return(false);
	}
	else if((str[0]>='0')&&(str[0]<='9')) switch(str[lcc])
	{
		case 'b': case 'o': case 'd': case 'h': break;
		default: lcc++;	//def: assume default base
	}
	else return(false);

	if(lcc < 1) return(false);
	for(;lcc;lcc--)				//get the things work
	{
		if( ((*str>='0')&&(*str<='9')) || ((*str>='A')&&(*str<='F')) ||
			((*str>='a')&&(*str<='f')) ) ;
		else return(false);
		str++;
	}

	return(true);
}


//-----------------------------------------------------------------------------
// INTMATH::extract: extracts a value (base=b/o/d/h) from a string
// default base is stored in [imo_def_base], returns IMRC
//-----------------------------------------------------------------------------
DWORD INTMATH::extract(char *str, __int64 *val)
{
	int		lcc = strlen(str)-1;	//points to last char
	__int8	base;					//the selected base
	__int8	digit;					//the current digit

	if(val) *val = NULL;			//start very low :)
	if(lcc < 0) return(RETC_PARAM_EXPECTED);

	if(str[0] == '$')
	{
		if(!enable_$) return(RETC_SYNTAX_ERR);
		base=IMO_BASE_HEX;
		str++;
		if(lcc < 1) return(RETC_SYNTAX_ERR);
	}
	else if((str[0] == '0')&&(str[1] == 'x'))
	{
		base=IMO_BASE_HEX;
		str+=2;
		if((--lcc) < 1) return(RETC_SYNTAX_ERR);
	}
	else if(str[0] == '\'')
	{
		if((!lcc)||(str[0]!='\'')) return(RETC_MARK_MISSING);
		digit=0;
		while(*(++str))
		{
			if(*str=='\'')
			{
				switch(*(++str))
				{
				case NULL: return(RETC_OK);
				case '\'': break;
				default: return(RETC_MARK_MISSING);
				}
			}
			if((++digit) > 8) return(RETC_OVERFLOW);
			if(val)	*val=(*val*256) + *str;

		}
		return(RETC_MARK_MISSING);
	}
	else if((str[0]>='0')&&(str[0]<='9')) switch(str[lcc])
	{
		case 'b': if(def_base==IMO_BASE_HEX) {base=IMO_BASE_HEX; lcc++;}
					else base=IMO_BASE_BIN; break;
		case 'o': base=IMO_BASE_OCT; break;
		case 'd': base=IMO_BASE_DEC; break;
		case 'h': base=IMO_BASE_HEX; break;
		default: base=def_base; lcc++;	//def: assume default base
	}
	else return(RETC_SYNTAX_ERR);

	if(lcc < 1) return(RETC_SYNTAX_ERR);

	BYTE cnt=0;
	for(;lcc;lcc--)				//get the things work
	{
		digit=0;
		switch(*(str++))
		{
		case 'F': case 'f': digit++; case 'E': case 'e': digit++;
		case 'D': case 'd': digit++; case 'C': case 'c': digit++;
		case 'B': case 'b': digit++; case 'A': case 'a': digit++;
		case '9': digit++; case '8': digit++;
		case '7': digit++; case '6': digit++; case '5': digit++; case '4': digit++;
		case '3': digit++; case '2': digit++; case '1': digit++; case '0': break;
		default: return(RETC_SYNTAX_ERR);
		}
		if(digit>=base) return(RETC_SYNTAX_ERR);
		if(val) *val = (*val * base) + digit;
		cnt++;
	}

	//overflow-check
	switch(base)
	{
	case IMO_BASE_BIN: if(cnt>64) return(RETC_OVERFLOW); break;
	case IMO_BASE_OCT: if(cnt>21) return(RETC_OVERFLOW); break;
	case IMO_BASE_DEC: if(cnt>19) return(RETC_OVERFLOW); break;
	case IMO_BASE_HEX: if(cnt>16) return(RETC_OVERFLOW); break;
	}

	return(RETC_OK);
}



//-----------------------------------------------------------------------------
int INTMATH::lookup_operator(char *s, int pos)
{
	int oc;
	int op_len;

	int last_op = _IMO_LAST;
	int last_len = 0;

	bool followed = true;
	switch(s[pos])
	{
	case ' ': case '\t':
	case '(': case 0: followed = false;
	}

	for(oc = _IMO_LAST; oc--;)
	{
		switch(oc)
		{
		case IMO_B_HI:
		case IMO_B_LO:
			if(followed) break;

		default:
			op_len = strlen(imo[oc]);
			if((op_len > last_len) && (pos >= op_len)) //is possible
			{
				if( assemble.cmp(imo[oc], &s[pos-op_len], op_len)) //found???
				{
				  last_op = oc;
				  last_len = op_len;
				}
			}
		}
	}

	return(last_op);
}
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// INTMATH::search_op: searches the input string for all op's in the imo
// (top2bottom), skipping brackets (right 2 left)
// returns true on success and fills RET_SEARCH, otherwise false on error
//-----------------------------------------------------------------------------
BOOL INTMATH::search_op(char *in, IMO_SEARCH_RET *sr)
{
	int pos;					//pos in string to search (right2left)

	int marks=0;				//is 1 inside of marks (")
	int brk_cntr=0;				//counts the opened brackets

	int last_op = _IMO_LAST;	//nothing found yet
	int last_pos;				//last position of opcode (number of first char)


	sr->prev=sr->rest=NULL;				//set 2 default:
	sr->op=-1;							//  no op found
	sr->rcode=RETC_OK;					//  no error


	for(pos = strlen(in); pos >= 0; pos--)	//scan string
	{
		switch(in[pos])
		{
		case '\'': marks=marks^1; break;
		case ')': if(!marks) brk_cntr++; break;
		case '(': if(!marks)
				  {
					if(brk_cntr) brk_cntr--;
					else
					{
					sr->rcode=RETC_BRACKET_OPEN;	//error: bracket not closed
					return(false);
					}
				  }
		}

		if((!brk_cntr)&&(!marks))	//skip things inside of those brackets&&marks
									//otherwise: search for operators
		{

			brk_cntr = lookup_operator(in, pos);

			if(imo_prio[brk_cntr] < imo_prio[last_op])	//found a better operator
			{
				last_op = brk_cntr;
				last_pos = pos;
				pos -= strlen(imo[brk_cntr]);
			}

			brk_cntr = 0;  //restore

		}
	}

	if(brk_cntr)	//terminate with open bracket??? -> error
	{
		sr->rcode=RETC_BRACKET_CLOSE;	//error: bracket not opened
		return(false);
	}
	if(marks)		//terminate with open string??? -> error
	{
		sr->rcode=RETC_MARK_MISSING;	//error: mark is missing
		return(false);
	}

	if(last_op < _IMO_LAST)
	{
		sr->op = last_op;
		brk_cntr = strlen(imo[last_op]);

		if(last_pos > brk_cntr)
		{
			sr->prev = in;
			in[last_pos - brk_cntr] = NULL;
		}

		if(in[last_pos]) sr->rest = &in[last_pos];

		return(true);
	}

	//done, but no op's found
	sr->rest=in;
	return(true);
}

//-----------------------------------------------------------------------------
// INTMATH::count_brk: counts the brackets in string in with len characters
//-----------------------------------------------------------------------------
int INTMATH::count_brk(char *in, int len)
{
	int marks=0;
	int brk_cntr=0;	//counts the opened brackets

	for(;len>0;len--)
	{
		switch(*(in++))
		{
		case '\'': marks=marks^1; break;
		case '(': if(!marks) {brk_cntr++;} break;
		case ')': if(!marks) {if(brk_cntr>0) brk_cntr--; else return(-1);}
		}
	}

	return(brk_cntr);
}

//-----------------------------------------------------------------------------
// INTMATH::do_operator: build return-value using operator, returns IMRC
//-----------------------------------------------------------------------------
DWORD INTMATH::do_operator(__int8 op, __int64 p1, __int64 *p2)
{
	#define NOT(a) (a)^(__int64)(-1)
	#define BOOL(a) ((a)?1:0)

	switch(op)
	{
	//LOGICAL operators
	case IMO_L_AND:		*p2 = ((p1)&&(*p2))?1:0; break;
	case IMO_L_OR:		*p2 = ((p1)||(*p2))?1:0; break;
	case IMO_L_EQ:		*p2 = ((p1)==(*p2))?1:0; break;
	case IMO_L_NEQ:		*p2 = ((p1)!=(*p2))?1:0; break;
	case IMO_L_GEQ:		*p2 = ((p1)>=(*p2))?1:0; break;
	case IMO_L_LEQ:		*p2 = ((p1)<=(*p2))?1:0; break;
	case IMO_L_GRT:		*p2 = ((p1)> (*p2))?1:0; break;
	case IMO_L_LWR:		*p2 = ((p1)< (*p2))?1:0; break;

	//MATH operators
	case IMO_M_ADD:		*p2 = p1 + *p2;			break;
	case IMO_M_SUB:		*p2 = p1 - *p2;			break;

	case IMO_M_MUL:		*p2 = p1 * *p2;			break;
	case IMO_M_DIV:
		if(*p2 == 0) {*p2 = NULL; return(RETC_DIV_BY_ZERO);}
		*p2 = p1 / *p2;
		break;
	case IMO_M_MOD:		*p2 = p1 % *p2;			break;

	//BTWISE operators
	case IMO_B_SHL:		*p2 = p1 << *p2;		break;
	case IMO_B_SHR:		*p2 = p1 >> *p2;		break;
	case IMO_B_AND:		*p2 = p1 & *p2;			break;
	case IMO_B_NAND:	*p2 = NOT(p1 & *p2);	break;
	case IMO_B_OR:		*p2 = p1 | *p2;			break;
	case IMO_B_NOR:		*p2 = NOT(p1 | *p2);	break;
	case IMO_B_XOR:		*p2 = p1 ^ *p2;			break;
	case IMO_B_NXOR:	*p2 = NOT(p1 ^ *p2);	break;
	case IMO_B_NOT:		*p2 = NOT(*p2);			break;

	case IMO_B_HI:	*p2 = (*p2 >> 16) & 0xFFFF; break;
	case IMO_B_LO:	*p2 = (*p2      ) & 0xFFFF;	break;

	default: *p2=NULL; return(RETC_SYNTAX_ERR);
	}

	return(RETC_OK);
}


//-----------------------------------------------------------------------------
// INTMATH::do_math: calculates the string passed to this rec-routine
// frees terms in brackets, calculates op's and returns the entire result
//-----------------------------------------------------------------------------
BOOL INTMATH::do_math(IM_RET *imr, char iflg)
{
	int lcc = strlen(imr->str)-1;	//offs of last char in string
	imr->imrc = RETC_OK;

	//skip spaces from right/left to center
	while( (lcc>=0) && ((imr->str[lcc]==' ')||(imr->str[lcc]=='\t')) )
	{
		imr->str[lcc]=NULL;
		lcc--;
	}
	while( (*imr->str) && ((*imr->str==' ')||(*imr->str=='\t')) )
	{
		*imr->str=NULL;
		imr->str++;
		lcc--;
	}

	//free term (if locked with brackets)
	if((lcc>0)&&(*imr->str=='(')&&(imr->str[lcc]==')')&&(!count_brk(imr->str+1, lcc-1)))
	{
		imr->str[lcc]=NULL;
		imr->str++;
		return( do_math(imr, iflg) );
	}

	__int64 param;	//save the first param before doing a seccond call to do_math (!!!)
	IMO_SEARCH_RET sr;
	if(search_op(imr->str, &sr))
	{	//search_op -> success
		if(sr.op == -1)	//no more operators found???
		{	//-> YES
			//found the exciting one, now figure it out (the isolated param)
			if((!sr.rest)||(!*sr.rest)) {imr->imrc=RETC_PARAM_EXPECTED; return(false);}

			#ifdef MATHDBG
				printf("%s", sr.rest);
			#endif

			//check for number
			if((imr->imrc = extract(sr.rest, &imr->value))!=RETC_OK)
			{	//otherwise check for label
				if(!label.validate(sr.rest)) return(false);
				imr->imrc = RETC_OK;
				LABELLIST *ll = label.get(sr.rest);
				if(!ll)		//does label exist already???
				{
					if((!(iflg&ITYPE_NOUNDEF))&&((assemble.preprocess)||(iflg&ITYPE_LOGICAL)))	//return zer0 for undefined labels
					{
						imr->value = 0;
						return(true);
					}

					imr->imrc = RETC_LABEL_UNDEF;
					return(false);
				}
				switch(ll->type)
				{
				case LBL_TYPE_REG:
					//imr->imrc = RETC_WRONG_PTYPE;	//regs not allowed
					//return(false);

				case LBL_TYPE_PC:
				case LBL_TYPE_PC_FOFS:
				case LBL_TYPE_EQUATE:
				case LBL_TYPE_DEFINE: break;  //always allow this definitions/constants

				default:
					if(iflg&ITYPE_NOLABELS)
					{
						imr->imrc = RETC_NO_LBLS_ALLOWED;
						return(false);
					}
				}

				//success
				imr->value = ll->value;

			}
			return(true);	//no processing yet
		}

		else

		{	//-> NO, here comes the operator
			//now test for asocciativity
			switch(imo_type[sr.op])
			{

			case IMO_ASSOC_R:	//no params before operator
				if(sr.prev)	{imr->imrc=RETC_TOO_MUCH_PARAMS; return(false);}
				if(!sr.rest) {imr->imrc=RETC_PARAM_EXPECTED; return(false);}

				#ifdef MATHDBG
					printf("%s(", imo[sr.op]);
				#endif
				imr->str=sr.rest;
				if(!do_math(imr, iflg)) return(false);
				#ifdef MATHDBG
					printf(")");
				#endif

				if((imr->imrc = do_operator(sr.op, 0, &imr->value))!=RETC_OK) return(false);
				return(true);

			case IMO_ASSOC_LR:	//params expected before and behind the op
				if((!sr.prev)||(!sr.rest)) {imr->imrc=RETC_PARAM_EXPECTED; return(false);}

				#ifdef MATHDBG
					printf("%s(", imo[sr.op]);
				#endif
				imr->str=sr.prev;
				if(!do_math(imr, iflg)) return(false);
				param=imr->value;
				#ifdef MATHDBG
					printf(",");
				#endif
				imr->str=sr.rest;
				if(!do_math(imr, iflg)) return(false);
				#ifdef MATHDBG
					printf(")");
				#endif
				if((imr->imrc = do_operator(sr.op, param, &imr->value))!=RETC_OK) return(false);
				return(true);

			case IMO_ASSOC_LDEF:	//params before op have default-value
				if(!sr.rest) {imr->imrc=RETC_PARAM_EXPECTED; return(false);}

				#ifdef MATHDBG
					printf("%s(", imo[sr.op]);
				#endif
				if(sr.prev)	//does 1st param exist???
				{	//-> YES, pick it up
					imr->str=sr.prev;
					if(!do_math(imr, iflg)) return(false);
					param=imr->value;
				}
				else
				{	//-> otherwise assume zero as default
					param=0;
					#ifdef MATHDBG
						printf("0");
					#endif
				}
				#ifdef MATHDBG
				  printf(",");
				#endif
				imr->str=sr.rest;
				if(!do_math(imr, iflg)) return(false);
				#ifdef MATHDBG
					printf(")");
				#endif

				if((imr->imrc = do_operator(sr.op, param, &imr->value))!=RETC_OK) return(false);
				return(true);

			}
		}
	}
	else
	{	//search_op -> error
		imr->imrc = sr.rcode;
	}

	return(false);	//error
}


//-----------------------------------------------------------------------------
// INTMATH::calc: the global caller 4 the internal math-routine
// (strings processed by this routine are not changed during call)
//-----------------------------------------------------------------------------
BOOL INTMATH::calc(IM_RET *imr, char iflg)
{
	if((!imr->str)||(!*imr->str)) {imr->imrc=RETC_PARAM_EXPECTED; return(false);}
	char *str = (char *)malloc(strlen(imr->str)+1);
	if(!str) do_error(ERR_FATAL|RETC_OUTOFMEM, ERR_SHOW_NOTHING);
	strcpy(str, imr->str);
	imr->str=str;

	#ifdef MATHDBG
	  printf("\"%s\" -> ", str);
	#endif

	BOOL b=do_math(imr, iflg);

	#ifdef MATHDBG
	  if(!b) printf(" -> error#%i\n", imr->imrc);
	  else printf(" -> 0x%X\n", imr->value);
	#endif

	free(str);
	imr->str=NULL;

	return(b);
}


//-----------------------------------------------------------------------------
// INTMATH::constructor
//-----------------------------------------------------------------------------
INTMATH::INTMATH()
{
	def_base = IMO_BASE_DEC;
	enable_$ = false;
}

//-----------------------------------------------------------------------------
// INTMATH::destructor
//-----------------------------------------------------------------------------
INTMATH::~INTMATH()
{
}


/*-----------------------------------------------------------------------------
// EXAMPLE
//-----------------------------------------------------------------------------
	char input[] = {"-1Fh * ( 37o - 01101b )"};
	INTMATH intmath;
	IM_RET imr;

	imr.str=input;
	if(intmath.calc(&imr))
	{
		printf("value = %i\n", imr.value);
	}
	else
	{
		printf("\nERROR(%i): ", imr.imrc);
	}
//---------------------------------------------------------------------------*/
