//=============================================================================
// STREAM CLASS
//=============================================================================

#include "streams.h"
#include "assemble.h"
#include "includes.h"

//-----------------------------------------------------------------------------
// STREAMin::line: reads the nxt line
//-----------------------------------------------------------------------------
char *STREAMin::line(BYTE mode)
{
    dbgout("STREAMin::line()\n")

    FILELIST *fl = stream.get();

    if(!fl)
    {
        do_error(ERR_FATAL|RETC_LISTREQUEST, 0);
    }

    if(fl->type&FTYPE_DONE) goto eof;	//already completed

    fl->line_cntr++;

    if(fl->type&FTYPE_BUF)
        //data is in a buffer -----------------------------------------------------
    {
        if((fl->pos == fl->buf) && (++stream.rec_cntr > MAX_REC))
        {
            do_error(RETC_TOO_MANY_CALLS, ERR_SHOW_POS);
            goto eof;
        }

        if((!fl->pos)||(fl->line_cntr > fl->size)) goto eof; //reached the end

        size_t len = strlen(fl->pos);
        if(len>=MAX_LINE-1) do_error(ERR_FATAL|RETC_LINE2LONG, ERR_SHOW_POS);
        strcpy(line_buf, fl->pos);
        fl->pos+=len+1;

        return(line_buf);
    } else {
    //otherwise: data is in a file --------------------------------------------
        if(!fl->fh)	//handle not initialized?
        {
            //count the opened files
            if((++stream.file_cntr) > MAX_FILES)
            {
                do_error(ERR_FATAL|RETC_TOO_MANY_FILES, ERR_SHOW_NOTHING);
            }



            if( !(fl->fh = fopen(fl->name, "rt")) )
            {
                if(!fl->last) do_error(ERR_FATAL|RETC_FOPEN, ERR_SHOW_FILE);
                //if it returns, give them a 2nd chance
                do_error(RETC_FOPEN, ERR_SHOW_FILE);
                goto eof;
            }
        }

        line_buf[MAX_LINE-2]=0;
        if(!fgets(line_buf, MAX_LINE, fl->fh))
            if(feof(fl->fh)) goto eof;	//reached end of file??? (line_buf not changed)
            else	if(fl->last) goto eof;
            else do_error(ERR_FATAL|RETC_FREAD, ERR_SHOW_FILE); //error
        else;
    }

    if(line_buf[MAX_LINE-2]) do_error(ERR_FATAL|RETC_LINE2LONG, ERR_SHOW_POS);
    return(line_buf);

    //reached the end of a file -----------------------------------------------
eof:
    dbgout("STREAMin::line() EOF\n")

        fl->line_cntr--;	//step back

    if(mode == STREAM_THIS_FILE) return(NULL);

    //do file-type depending stuff
    switch(fl->type&FTYPE_MASK)
    {
        case FTYPE_DEFAULT:
            fl->type = fl->type | FTYPE_DONE;           //file done
            stream.file_no = fl->last;                  //switch back to previous handle
            break;

        case FTYPE_INCLUDE:
            if(!assemble.preprocess)
            {
                // if(ferr) fprintf(ferr, "%s(%i) - %i errors, %i warnings\n", fl->name, fl->line_cntr, fl->error_cntr, fl->warning_cntr);
                // else printf("%s(%i) - %i errors, %i warnings\n", fl->name, fl->line_cntr, fl->error_cntr, fl->warning_cntr);
            }

            fl->type = fl->type | FTYPE_DONE;	//file done
            stream.file_no = fl->last;	//switch back to previous handle

            if((fl->error_cntr)||(fl->warning_cntr))
            {
                if(fl->line_cntr) do_error(ERR_WARNING|RETC_INCLUDE_ERR, ERR_SHOW_POS);
                else do_error(ERR_WARNING|RETC_INCLUDE_ABORT, ERR_SHOW_POS);
            }
            break;

        case FTYPE_MACRO:
            fl->type = fl->type | FTYPE_DONE;	//file done
            stream.file_no = fl->last;	//switch back to previous handle
            if((fl->error_cntr)||(fl->warning_cntr))
            {
                if(fl->pos == fl->buf) do_error(ERR_WARNING|RETC_MACRO_ABORT, ERR_SHOW_POS);
                else do_error(ERR_WARNING|RETC_MACRO_ERR, ERR_SHOW_POS);
            }
            break;

        case FTYPE_IFELSE:
            fl->type = fl->type | FTYPE_DONE;	//file done
            stream.file_no = fl->last;	//switch back to previous handle
            if((fl->error_cntr)||(fl->warning_cntr))
            {
                if(fl->pos == fl->buf) do_error(ERR_WARNING|RETC_IFELSE_ABORT, ERR_SHOW_POS);
                else do_error(ERR_WARNING|RETC_IFELSE_ERR, ERR_SHOW_POS);
            }
            break;

        default:
            stream.file_no = fl->last;	//switch back to previous handle
    }

    //if buf then free it, else close file and dec open-file-cntr
    if(fl->type&FTYPE_BUF)
    {
        stream.rec_cntr--;
        if(fl->buf) free(fl->buf);
        fl->buf = NULL;
    }
    else
    {
        stream.file_cntr--;
        if(fl->fh) fclose(fl->fh);
        fl->fh = NULL;
    }

    if(!stream.file_no) return(NULL);	//handle to last file??? then return with error

    return(line());
}

//-----------------------------------------------------------------------------
// STREAMin::constructor
//-----------------------------------------------------------------------------
STREAMin::STREAMin()
{
    line_buf = (char *)malloc(MAX_LINE);
    if(!line_buf) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
}

//-----------------------------------------------------------------------------
// STREAMin::destructor
//-----------------------------------------------------------------------------
STREAMin::~STREAMin()
{
    free(line_buf);
}




//-----------------------------------------------------------------------------
// STREAMout::fopen: create file 4 output
//-----------------------------------------------------------------------------
void STREAMout::open()
{
    FILELIST *fl = stream.get();
    if(!fl) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);

    if( !fl->name ) fh = NULL;
    else if( !(fh = fopen(fl->name, "wb")) )
        do_error(ERR_FATAL|RETC_FCREATE, ERR_SHOW_FILE);

    fofs = NULL;
}

//-----------------------------------------------------------------------------
// STREAMout::file: writes entire selected file 2 binout
//-----------------------------------------------------------------------------
DWORD STREAMout::file()
{
    dbgout("STREAMout::file()\n")

        FILELIST *fl = stream.get();	//get file 2 stream out
    if(!fl) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);

    //open "fl->name", read in allocated buffer and stream out, close
    if( !(fl->fh = fopen(fl->name, "rb")) )
    {
        do_error(RETC_FOPEN, ERR_SHOW_FILE);
        stream.file_no = fl->last;
        return(ERR_WARNING|RETC_INCLUDE_ERR);
    }

    flush();

    if(fh)	//writing allowed?
    {
        while(!feof(fl->fh))
        {
            wrt_pos = fread(wrt_buf, 1, 1024, fl->fh);
            if(wrt_pos)
            {
                assemble.prog_cntr += wrt_pos;
                fofs += wrt_pos;
                flush();
            }
            if(ferror(fl->fh))
            {
                do_error(RETC_FREAD, ERR_SHOW_FILE);
                fclose(fl->fh);
                fl->fh=NULL;
                stream.file_no = fl->last;
                return(ERR_WARNING|RETC_INCLUDE_ABORT);
            }
        }
    } else	{
        fpos_t fsize;

        //in this case, just get filesize and add to pointers
        if((fseek(fl->fh, NULL, SEEK_END)) || (fgetpos(fl->fh, &fsize)))
        {
            do_error(RETC_FREAD, ERR_SHOW_FILE);
            fclose(fl->fh);
            fl->fh=NULL;
            return(ERR_WARNING|RETC_INCLUDE_ABORT);
        }

        assemble.prog_cntr += fsize;
        fofs += fsize;
    }

    fclose(fl->fh);
    fl->fh=NULL;
    stream.file_no = fl->last;	//return 2 last file

#ifdef _DEBUG
    dbgout2("<file> ")
#endif
    return(RETC_OK);
}

//-----------------------------------------------------------------------------
// STREAMout::buf: stores count bytes from buffer
//-----------------------------------------------------------------------------
void STREAMout::buf(char *buf, WORD count)
{
    if(fh)
    {
        if(fwrite(buf, 1, count, fh) != (size_t)count)
            do_error(ERR_FATAL|RETC_FWRITE, ERR_SHOW_NOTHING);
    }

    assemble.prog_cntr += count;
    fofs += count;
}

//-----------------------------------------------------------------------------
// STREAMout::opcode: stores value ([count] bytes) to output-buf
//-----------------------------------------------------------------------------
void STREAMout::opcode(__int64 value, WORD count)
{
#ifdef _DEBUG
    if(!assemble.preprocess) for(int i=count;i--;) dbgout2((int)((BYTE*)&value)[i] << " ")
#endif

        if(count>8)	//__int64 holds 8 bytes long
        {
            do_error(ERR_WARNING|RETC_OVERFLOW, ERR_SHOW_POS);
            opcode(value, 8);
            align(count);
            return;
        }

    //check overflow
    //	if( ((unsigned __int64)value) & (((((unsigned __int64)1) << (count << 3))-1)^((unsigned __int64)-1)) )
    //		do_error(ERR_WARNING|RETC_OVERFLOW, ERR_SHOW_POS);

    if(fh)
    {
        if(wrt_pos+count >= WRITE_BUF) flush();
        memcpy(wrt_buf+wrt_pos, &value, count);
        wrt_pos += count;
    }

    assemble.prog_cntr += count;
    fofs += count;
}

//-----------------------------------------------------------------------------
// STREAMout::align: increases the code-cntr to a multiple of [align]
//-----------------------------------------------------------------------------
void STREAMout::align(__int64 align)
{
    while(assemble.prog_cntr % align) opcode(NULL, 1);
}

//-----------------------------------------------------------------------------
// STREAMout::flush: writes all the buffer 2 the file
//-----------------------------------------------------------------------------
void STREAMout::flush()
{
    if(!wrt_pos) return;

    if(fh)
    {
        if(fwrite(wrt_buf, 1, wrt_pos, fh) != (size_t)wrt_pos)
            do_error(ERR_FATAL|RETC_FWRITE, ERR_SHOW_NOTHING);
    }

    wrt_pos = 0;
}

//-----------------------------------------------------------------------------
// STREAMout::close file
//-----------------------------------------------------------------------------
void STREAMout::close()
{
    if(!fh) return;

    flush();
    fclose(fh);
    fh = NULL;
}

//-----------------------------------------------------------------------------
// STREAMout::constructor
//-----------------------------------------------------------------------------
STREAMout::STREAMout()
{
    wrt_buf = (char *)malloc(WRITE_BUF);
    if(!wrt_buf) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
}

//-----------------------------------------------------------------------------
// STREAMout::destructor
//-----------------------------------------------------------------------------
STREAMout::~STREAMout()
{
    close();
    free(wrt_buf);
}




//-----------------------------------------------------------------------------
// STREAM::add: add file 2 filelist
//-----------------------------------------------------------------------------
FILELIST *STREAM::add(char *fname, BYTE type)
{
    dbgout("STREAM::add(CHAR*[" << (DWORD)fname << "], INT[" << (int)type << "])\n")
        FILELIST *fl;

    fl = (FILELIST*)FileList->add();
    if(!fl) do_error(ERR_FATAL|RETC_LISTALLOC, 0);

    fl->type = type;	//FTYPE_xxx
    if(file_no>0)
    {
        fl->last = file_no;
    }

    //fl->line_cntr = 0;
    //fl->error_cntr = 0;
    //fl->warning_cntr = 0;

    if(fname)
    {
        fl->name = (char *)malloc(strlen(fname)+1);
        if(!fl->name) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
        strcpy(fl->name, fname);
    }

    file_no = FileList->members-1;

    return(fl);
}

//-----------------------------------------------------------------------------
// STREAM::get: get file-info of actual file
//-----------------------------------------------------------------------------
FILELIST *STREAM::get()
{
    if(file_no == -1) return(NULL);
    return((FILELIST *)FileList->get(file_no));
}

//-----------------------------------------------------------------------------
// STREAM::get: get file-info of actual file
//-----------------------------------------------------------------------------
FILELIST *STREAM::get(DWORD no)
{
    return((FILELIST *)FileList->get(no));
}

//-----------------------------------------------------------------------------
// STREAM::close: free strings in list and give error hints
//-----------------------------------------------------------------------------
void STREAM::close()
{
    // free strings in FileList
    FILELIST *fl;
    for(DWORD cnt = FileList->members; cnt; )
        if(fl = (FILELIST *)FileList->get(--cnt))
        {
            if(fl->name) free(fl->name);
            fl->name = NULL;

            if(fl->type&FTYPE_BUF)
            {
                if(fl->buf) free(fl->buf);
                fl->buf = NULL;
            }
            else
            {
                if(fl->fh) fclose(fl->fh);
                fl->fh = NULL;
            }
        }

    FileList->kill();		//reset list

    file_no = (DWORD) - 1;			//the actual file for input from the FileList
    file_cntr = 0;			//number of files open at once
    rec_cntr = 0;			//number of open buffers (recursive calls)
}

//-----------------------------------------------------------------------------
// STREAM::constructor
//-----------------------------------------------------------------------------
STREAM::STREAM()
{
    dbgout("STREAM::STREAM()\n")

    FileList = new LIST( sizeof(FILELIST) );

    file_no = (DWORD) - 1;			//the actual file for input from the FileList
    file_cntr = 0;			//number of files open at once
    rec_cntr = 0;			//number of open buffers (recursive calls)
}

//-----------------------------------------------------------------------------
// STREAM::destructor
//-----------------------------------------------------------------------------
STREAM::~STREAM()
{
    dbgout("STREAM::~STREAM()\n")
    close();
    delete FileList;

#ifndef BSD
    _fcloseall();
#endif
}
