//=============================================================================
// ASMCLASS CLASS
//=============================================================================

#include "asmclasses.h"

#include "labels.h"
#include "assemble.h"
#include "streams.h"


//-----------------------------------------------------------------------------
// ASMCLASS::_get: search for a special class
//-----------------------------------------------------------------------------
DWORD ASMCLASS::_get(char *name, DWORD last)
{
    dbgout("ASMCLASS::_get(CHAR[" << name << "], DWORD[" << last << "]) -> ")

	CLASSLIST *cl;
	for(DWORD cnt = ClassList->members; cnt; )
	if(cl = (CLASSLIST *)ClassList->get(--cnt))
	{
		if((assemble.cmp(cl->name, name)) && (cl->last == last))
		{
            dbgout(cnt << "\n")
			return(cnt);
		}
	}

    dbgout("not_found\n")
	return((DWORD)-1);
}

//-----------------------------------------------------------------------------
// ASMCLASS::get: get a actual opened seg/proc
//-----------------------------------------------------------------------------
CLASSLIST *ASMCLASS::get(DWORD no)
{
	dbgout("ASMCLASS::get(DWORD[" << no << "]) ")

    if(no == (DWORD) - 1)
    {
        if(sel == (DWORD) - 1)
        {
            return(NULL);
        }
        no = sel;
    }

	CLASSLIST *cl = (CLASSLIST *)ClassList->get(no);
	if(!cl)
    {
        do_error(ERR_FATAL|RETC_LISTREQUEST, 0);
    }
	return(cl);
}

//-----------------------------------------------------------------------------
// ASMCLASS::add: add segment or proc 2 list
//-----------------------------------------------------------------------------
DWORD ASMCLASS::add(char *name, __int8 type)
{
	CLASSLIST *cl;
	if((!name)||(!*name)) return(RETC_PARAM_EXPECTED);

    dbgout("ASMCLASS::add(CHAR[" << name << "], INT8[" << (WORD)type << "])\n")

	if(!label._validate(name)) return(RETC_UNKWN_LBL_SYNTAX);

    switch(type)
    {
        case CLS_TYPE_SEG:
            if(sel != -1) close_all();
            break;

            //case CLS_TYPE_PROC:
            //case CLS_TYPE_TBL:
    }

	if(!label.add(name, assemble.prog_cntr, LBL_TYPE_CODE, _LOCAL_LBL))
    {
		return(RETC_OK);	//label has alreaddy give an error
    }

	//search, if already defined
	DWORD nxt_sel;
	if((nxt_sel = _get(name, sel)) != -1)
	{	//overwrite!!
		cl = get(nxt_sel);
		if(cl->pass == assemble.pass)
        {
            return(RETC_LBL_OVERWRITE);
        }
	} else {	//new!!
		cl = (CLASSLIST *)ClassList->add();
		if(!cl) do_error(ERR_FATAL|RETC_LISTALLOC, 0);

		nxt_sel = ClassList->members - 1;

		cl->name = (char *)malloc(strlen(name)+1);
		if(!cl->name) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
		strcpy(cl->name, name);
	}

	//get info about actual file (line_cntr, only if available)
	FILELIST *fl = stream.get();
	cl->file_no = stream.file_no;
	if(fl)
    {
        cl->line_cntr = fl->line_cntr;
    }

	cl->type = type;
	cl->last = sel;

	cl->pass = assemble.pass;

	cl->start = assemble.prog_cntr;
	cl->fstart = stream.out.fofs;

	sel = nxt_sel;

	if(assemble.binary)
    {
        cl->_fsize = label.add("_fsize", 0, LBL_TYPE_PREDEFINE, _LOCAL_LBL);
    } else {
        cl->_fsize = NULL;
    }
	cl->_size = label.add("_size", 0, LBL_TYPE_PREDEFINE, _LOCAL_LBL);
	if(assemble.binary)
    {
        label.add("_fofs", cl->fstart, LBL_TYPE_UNDEF, _LOCAL_LBL);
    }
	//else label.add("_fofs", cl->fstart, LBL_TYPE_PREDEFINE, _LOCAL_LBL);
	label.add("_ofs", cl->start, LBL_TYPE_UNDEF, _LOCAL_LBL);

	return(RETC_OK);
}

//-----------------------------------------------------------------------------
// ASMCLASS::close: set close-bit in the last seg/proc-listmember
//-----------------------------------------------------------------------------
DWORD ASMCLASS::close(char *name, __int8 type)
{
    dbgout("ASMCLASS::close(CHAR[" << name << "], INT8[" << (WORD)type << "])\n")

	//close class and update vars
	CLASSLIST *cl = get();
	if((!cl) || ((name) && (!assemble.cmp(name, cl->name))) ) switch(type)
	{
	case CLS_TYPE_PROC: return(RETC_PROC_EXPECTED);
	case CLS_TYPE_TBL: return(RETC_TABLE_EXPECTED);
	case CLS_TYPE_SEG: return(RETC_SEG_EXPECTED);
	case CLS_TYPE_VPU: return(RETC_VPUCODE_EXPECTED);
	default: return((DWORD) - 1);
	}

	//cl->type = cl->type | CLS_TYPE_CLOSED;
	cl->end = assemble.prog_cntr;
	cl->fend = stream.out.fofs;

	label.preset(cl->end - cl->start, LBL_TYPE_UNDEF, cl->_size);
	if((assemble.binary)&&(cl->_fsize)) label.preset(cl->fend - cl->fstart, LBL_TYPE_UNDEF, cl->_fsize);

	sel = cl->last;

	return(RETC_OK);
}

//-----------------------------------------------------------------------------
// SEGMENT::close_all
//-----------------------------------------------------------------------------
void ASMCLASS::close_all()
{
    dbgout("ASMCLASS::close_all()\n")

	//check sel, and close
	DWORD ret;
	CLASSLIST *cl;
    while( cl = get() )
    {
        switch(cl->type)
        {
        case CLS_TYPE_PROC: do_error(RETC_ENDP_EXPECTED, ERR_SHOW_POS); break;
        case CLS_TYPE_TBL: do_error(RETC_ENDT_EXPECTED, ERR_SHOW_POS); break;
        case CLS_TYPE_SEG: do_error(RETC_ENDS_EXPECTED, ERR_SHOW_POS); break;
        case CLS_TYPE_VPU: do_error(RETC_ENDVPU_EXPECTED, ERR_SHOW_POS); break;
        default: do_error((DWORD)-1, ERR_SHOW_POS);
        }
        if((ret = close(NULL, cl->type)) != RETC_OK) do_error(ret, ERR_SHOW_POS);
    }
}

//-----------------------------------------------------------------------------
// SEGMENT::constructor
//-----------------------------------------------------------------------------
ASMCLASS::ASMCLASS()
{
    dbgout("ASMCLASS::ASMCLASS()\n")

	sel = (DWORD) - 1;

	ClassList = new LIST( sizeof(CLASSLIST) );
}

//-----------------------------------------------------------------------------
// SEGMENT::destructor
//-----------------------------------------------------------------------------
ASMCLASS::~ASMCLASS()
{
    dbgout("ASMCLASS::~ASMCLASS()\n")
    dbgout("\nclasslist:\n")

	CLASSLIST *cl;
	for(DWORD cnt = ClassList->members; cnt; )
	if(cl = (CLASSLIST *)ClassList->get(--cnt))
	{
        dbgout("#" << cnt << "\t" << cl->name << "\t" << (__int32)cl->last << "\n")
		if(cl->name) free(cl->name);
	}
    dbgout("\n")

	delete ClassList;
}
