#ifndef _INTMATH_
#define _INTMATH_


//-----------------------------------------------------------------------------
//---= include =---------------------------------------------------------------
//-----------------------------------------------------------------------------
#include "includes.h"
#include "errors.h"


//-----------------------------------------------------------------------------
//---= enum/define =-----------------------------------------------------------
//-----------------------------------------------------------------------------
enum						//all operators: ascending priority
{
	//LOGICAL operators
	IMO_L_AND,	//&&
	IMO_L_OR,	//||
	IMO_L_EQ,	//==
	IMO_L_NEQ,	//!=
	IMO_L_GEQ,	//>=
	IMO_L_LEQ,	//<=
	IMO_L_GRT,	//>
	IMO_L_LWR,	//<

	//MATH operators
	IMO_M_ADD,	//+
	IMO_M_SUB,	//-
	IMO_M_MUL,	//*
	IMO_M_DIV,	///
	IMO_M_MOD,	//%

	//BITWISE operators
	IMO_B_SHL,	//<<
	IMO_B_SHR,	//>>
	IMO_B_AND,	//&
	IMO_B_NAND,	//!&
	IMO_B_OR,	//|
	IMO_B_NOR,	//!|
	IMO_B_XOR,	//^
	IMO_B_NXOR,	//!^
	IMO_B_NOT,	//!

	IMO_B_HI,	//%HI
	IMO_B_LO,	//%LO

	_IMO_LAST
};

enum //Associativity
{
	IMO_ASSOC_LR,	//left to right
	IMO_ASSOC_R,	//right param
	IMO_ASSOC_LDEF,	//left param has a default value

	_IMO_ASSOC_LAST
};

enum
{
	ITYPE_DEFAULT,
	ITYPE_NOLABELS = 1,
	ITYPE_LOGICAL = 2,
	ITYPE_NOUNDEF = 4,
};

#define IMO_BASE_BIN 2		//bases for extracting/destilling numbers
#define IMO_BASE_OCT 8
#define IMO_BASE_DEC 10
#define IMO_BASE_HEX 16


//-----------------------------------------------------------------------------
//---= typedef =---------------------------------------------------------------
//-----------------------------------------------------------------------------
typedef struct				//do_math parameter block
{
	char	*str;			//points 2 input on call, and 2 error-part on return (sometimes :)
	__int64	value;
	DWORD	imrc;
} IM_RET;

typedef struct
{
	char	*prev;
	int		op;
	char	*rest;
	DWORD	rcode;
} IMO_SEARCH_RET;


//-----------------------------------------------------------------------------
//---= class =-----------------------------------------------------------------
//-----------------------------------------------------------------------------
class INTMATH
{
public:
	INTMATH();
	BOOL is_value(char *);
	DWORD extract(char *, __int64 *);
	BOOL calc(IM_RET *, char = ITYPE_DEFAULT);
	~INTMATH();

	__int8 def_base;
	BOOL enable_$;

private:
	int lookup_operator(char *s, int pos);
	BOOL search_op(char *, IMO_SEARCH_RET *);
	int count_brk(char *, int);
	DWORD do_operator(__int8, __int64, __int64 *);
	BOOL do_math(IM_RET *, char);
};


//-----------------------------------------------------------------------------
//---= extern =----------------------------------------------------------------
//-----------------------------------------------------------------------------
extern INTMATH intmath;
#endif
