//=============================================================================
// LABEL CLASS
//=============================================================================

#include "labels.h"

#include "intmath.h"
#include "asmclasses.h"
#include "assemble.h"
#include "streams.h"


//-----------------------------------------------------------------------------
// LABEL::validate_label: returns true if label is valid (allowed characters)
//-----------------------------------------------------------------------------
BOOL LABEL::validate(char *lbl)
{
	if(*lbl=='@') lbl++;

	for(;*lbl;lbl++)
		if(( (*lbl>='a')&&(*lbl<='z') ) ||
		   ( (*lbl>='A')&&(*lbl<='Z') ) ||
		   ( (*lbl>='0')&&(*lbl<='9') )    );
		else switch(*lbl)
		{case '~':case '.': case '_':case '-':case '$':break; default:return(false);}

	return(true);
}

//-----------------------------------------------------------------------------
// LABEL::_validate: does not allow operators '.','~','@'
//-----------------------------------------------------------------------------
BOOL LABEL::_validate(char *lbl)
{
	for(;*lbl;lbl++)
		if(( (*lbl>='a')&&(*lbl<='z') ) ||
		   ( (*lbl>='A')&&(*lbl<='Z') ) ||
		   ( (*lbl>='0')&&(*lbl<='9') )    );
		else switch(*lbl)
		{case '_':case '-':case '$':break; default:return(false);}

	return(true);
}

//-----------------------------------------------------------------------------
// LABEL::_get: search label, additional comparing class_numbers
//-----------------------------------------------------------------------------
DWORD LABEL::_get(char *lbl, DWORD cls)
{
	LABELLIST *ll;
	DWORD pos = LabelList->members;

	dbgout("LABEL::_get(CHAR[" << lbl << "], DWORD[" << cls << "]) ")

	if((lbl)&&(*lbl)) while(pos--)
	{
		ll = (LABELLIST *)LabelList->get(pos);
		if(!ll) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);

		if((ll->sel_class == cls) && (ll->type != LBL_TYPE_PREDEFINE) && (assemble.cmp(lbl, ll->name)))
		{
			switch(ll->type)
			{
			case LBL_TYPE_PC: ll->_val = assemble.prog_cntr; break;	//update label value of PC_LBL
			case LBL_TYPE_PC_FOFS: ll->_val = stream.out.fofs; break;
			}

			ll->value = ll->_val;
			dbgout("-> " << pos << "\n")
			return(pos);
		}
	}

	dbgout("-> not_found\n")
	return (DWORD)-1;
}

//-----------------------------------------------------------------------------
// LABEL::get: search lbltbl for lbl and return list-id
//-----------------------------------------------------------------------------
LABELLIST *LABEL::get(char *_lbl)
{
	CLASSLIST *cl;
	LABELLIST *ll;

	BOOL rel;
	__int64 _rel;

	DWORD sel_class;
	DWORD lbl_class;

	char *lbl, *nxt;

	dbgout("LABEL::get(CHAR[" << _lbl << "]) ")

	if(!validate(_lbl)) return(NULL);

	//now search locals
	if(assemble.config.default_local)
	{
		//if(*_lbl == '@') {sel_class = -1; _lbl++;}
		//else sel_class = asmclass.sel;
		if(*_lbl == '@') _lbl++;
		sel_class = asmclass.sel;	//by default
	}
	else
	{
		if(*_lbl == '@') {sel_class = asmclass.sel; _lbl++;}
		else sel_class = (DWORD)-1;
	}

search_again:	//here we meet again

	lbl_class = sel_class;
	lbl = _lbl;
	rel = false;

	cl = asmclass.get(lbl_class);
	if(cl) _rel = cl->start; else _rel = 0;

	//skip through classes
	while(nxt = strchr(lbl, '.'))
	{
		*nxt = NULL;
		if(*lbl == '~')
		{
			if(rel)	return(NULL);	//syntax error

			lbl++;
			rel = true;
			if((lbl_class = asmclass._get(lbl, lbl_class)) == -1)
			{
				*nxt = '.';
				goto next_class;
			}
		}
		else if((lbl_class = asmclass._get(lbl, lbl_class)) == -1)
		{
			*nxt = '.';
			goto next_class;
		}

		if(!rel)	//remember last class, until it is the relative one
		{
			cl = asmclass.get(lbl_class);
			if(cl) _rel = cl->start; else _rel = 0;
		}

		*nxt = '.';
		lbl = ++nxt;
	}

	if(*lbl == '~')
	{
		if(rel) return(NULL);	//syntax error

		rel = true;
		lbl++;
	}

	if((lbl_class = _get(lbl, lbl_class)) == -1) goto next_class;

	ll = (LABELLIST *)LabelList->get(lbl_class);
	if(!ll) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);

	if(rel) ll->value = ll->_val - _rel;
	else ll->value = ll->_val;

	return(ll);
	//////////

next_class:	//switch a class back (if the last searched class is the last, return NULL)

	if(sel_class == -1) return(NULL);
	cl = asmclass.get(sel_class);
	if(!cl) return(NULL);
	sel_class = cl->last;

	goto search_again;	// and search again
}


//-----------------------------------------------------------------------------
// LABEL::add: check if valid (no other defs), then add2tbl
//-----------------------------------------------------------------------------
DWORD LABEL::add(char *lbl, __int64 value, BYTE type, __int8 is_local)
{
	LABELLIST *ll;
	FILELIST *fl;

	dbgout("LABEL::add(CHAR[" << lbl << "], INT64[" << (DWORD)value << "], DWORD[" << (WORD)type << "], INT8[" << (WORD)is_local << "])\n")

	//detect global/local type
	DWORD sel_class;
	if(is_local != _UNDEF_LBL) sel_class = is_local;
	else
	{
		sel_class = assemble.config.default_local;
		if(*lbl == '@') switch(type)
		{
		case LBL_TYPE_PC:		//not allowed on this types
		case LBL_TYPE_PC_FOFS:
		case LBL_TYPE_REG:
		case LBL_TYPE_EQUATE:
		case LBL_TYPE_DEFINE:
			do_error(RETC_UNKWN_LBL_SYNTAX, ERR_SHOW_POS);
			return(false);

		default:
			lbl++;
			sel_class = sel_class^1; //swap the default mode on a @-operator
		}
		else switch(type)
		{
		case LBL_TYPE_PC:		//on this types always global
		case LBL_TYPE_PC_FOFS:
		case LBL_TYPE_REG:
		case LBL_TYPE_EQUATE:
		case LBL_TYPE_DEFINE:
			sel_class = 0;
		}
	}

	if(sel_class) sel_class = asmclass.sel;	//local label
	else sel_class = (DWORD)-1;	//global label

	//check label with internal-math-routine (should be no number)
	//and instruction routine (should be no label)
	//error, if overwrite
	if(intmath.is_value(lbl))
	{
		do_error(RETC_KEY_OVERWRITE, ERR_SHOW_POS);
		return(false);
	}

	//check all characters for valid range
	if(!_validate(lbl))
	{
		do_error(RETC_UNKWN_LBL_SYNTAX, ERR_SHOW_POS);
		return(false);
	}

	//on REGDEF check the range
	if((type == LBL_TYPE_REG)&&((value<0)||(value>31)))
	{
		do_error(RETC_REG_OVRFLW, ERR_SHOW_POS);
		return(false);
	}

	//search (by class), if already defined...
	DWORD lbl_no;
	if((lbl_no = _get(lbl, sel_class)) != -1)
	{
		//get existing label
		ll = (LABELLIST *)LabelList->get(lbl_no++);
		if(!ll) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);

		if(ll->pass == assemble.pass)
		{
			do_error(RETC_LBL_OVERWRITE, ERR_SHOW_POS);
			return(false);
		}

		//check for changes (new value ?= old value)
		if((type != LBL_TYPE_PREDEFINE)&&(ll->_val != value))
		{
			//if(!assemble.preprocess) printf("-> \"%s\"\n", lbl);
			assemble.pass_dependent = true;
		}
	}
	else
	{
		//add new label
		ll = (LABELLIST *)LabelList->add();
		if(!ll) do_error(ERR_FATAL|RETC_LISTALLOC, 0);

		lbl_no = LabelList->members;

		//if(!assemble.preprocess) printf("-> \"%s\"\n", lbl);
		assemble.pass_dependent = true;	//needs a new pass to check label
	}

	//get info about actual file (line_cntr, only if available)
	fl = stream.get();
	ll->file_no = stream.file_no;
	if(fl) ll->line_cntr = fl->line_cntr;

	//fill label info
	ll->type = type;
	ll->value = ll->_val = value;
	ll->sel_class = sel_class;
	ll->pass = assemble.pass;

	if(!ll->name)
	{
		ll->name = (char *)malloc(strlen(lbl)+1);
		if(!ll->name) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
		strcpy(ll->name, lbl);
	}

	//no warn, because every local lbl is different to all globals
	//if((sel_class != -1)&&(_get(lbl, -1)))
	//	do_error(ERR_WARNING|RETC_LOCAL_OVRWR_GLBL, ERR_SHOW_POS);

	//write listing
	if((!assemble.preprocess)&&(assemble.lst_fh)) switch(ll->type)
  {
  case LBL_TYPE_CODE:		//dont log these
	case LBL_TYPE_DB:
  case LBL_TYPE_DH:
	case LBL_TYPE_DW:
	case LBL_TYPE_DD:
	case LBL_TYPE_DQ:
#ifdef __GNUC__
    fprintf(assemble.lst_fh, "%08llX\t%08llX\t\t[%s]\n",
#else
    fprintf(assemble.lst_fh, "%08I64X\t%08I64X\t\t[%s]\n",
#endif
			assemble.prog_cntr, ll->value, ll->name);
  }

	dbgout("LABEL::add(" << lbl << ") -> TRUE\n")
	return(lbl_no);
}

//-----------------------------------------------------------------------------
// LABEL::predefine: add a predefined label (not valid untill set)
//-----------------------------------------------------------------------------
DWORD LABEL::predef(char *lbl)
{
	dbgout("LABEL::predefine(CHAR[" << lbl << "])\n")

	if(predef_lbl)
	{
		preset(assemble.prog_cntr, LBL_TYPE_UNDEF);
	}

	return( predef_lbl = add(lbl, 0, LBL_TYPE_PREDEFINE) );
}

//-----------------------------------------------------------------------------
// LABEL::if_predef
//-----------------------------------------------------------------------------
BOOL LABEL::if_predef()
{
	return(predef_lbl?true:false);
}

//-----------------------------------------------------------------------------
// LABEL::preset: define the last predefined label
//-----------------------------------------------------------------------------
BOOL LABEL::preset(__int64 value, BYTE type, DWORD lbl)
{
	dbgout("LABEL::preset(INT64[" << (DWORD)value << "], BYTE[" << (WORD)type << "])\n")

	if(!lbl)
	{
		if(!predef_lbl) return(false);
		lbl = predef_lbl;
		predef_lbl = 0;
	}

	if(!lbl) return(false);

	LABELLIST *ll = (LABELLIST *)LabelList->get(--lbl);
	if(!ll) do_error(ERR_FATAL|RETC_LISTALLOC, 0);
	if(ll->type != LBL_TYPE_PREDEFINE) return(false);

	ll->type = type;
	ll->value = ll->_val = value;
	ll->pass = assemble.pass;
	if(ll->_val != value)
	{
		//if(!assemble.preprocess) printf("-> \"%s\"\n", lbl);
		assemble.pass_dependent = true;
	}

	//write listing
	if((!assemble.preprocess)&&(assemble.lst_fh)) switch(ll->type)
  {
  case LBL_TYPE_CODE:		//dont log these
	case LBL_TYPE_DB:
  case LBL_TYPE_DH:
	case LBL_TYPE_DW:
	case LBL_TYPE_DD:
	case LBL_TYPE_DQ:
#ifdef __GNUC__
    fprintf(assemble.lst_fh, "%08llX\t%08llX\t\t[%s]\n",
#else
    fprintf(assemble.lst_fh, "%08I64X\t%08I64X\t\t[%s]\n",
#endif
			assemble.prog_cntr, ll->value, ll->name);
  }

	return(true);
}

//-----------------------------------------------------------------------------
// LABEL::constructor
//-----------------------------------------------------------------------------
LABEL::LABEL()
{
	dbgout("LABEL::LABEL()\n")

	LabelList = new LIST( sizeof(LABELLIST) );
	predef_lbl = 0;

	//init the label list
	add("$", NULL, LBL_TYPE_PC);	//add pc-label (which returns always prog_cntr)
	add("$$", NULL, LBL_TYPE_PC_FOFS);

	add("STACK", 0, LBL_TYPE_PREDEFINE);
	add("BSS", 0, LBL_TYPE_PREDEFINE);

	add("TRUE", -1, LBL_TYPE_DEFINE);	//add logical defines
	add("FALSE", 0, LBL_TYPE_DEFINE);
}

//-----------------------------------------------------------------------------
// LABEL::destructor
//-----------------------------------------------------------------------------
LABEL::~LABEL()
{
	dbgout("LABEL::~LABEL()\n")
	dbgout("\nlabellist:\n")

	DWORD cnt;
	LABELLIST *ll;


	//free LabelList (allocated names)
	for(cnt = LabelList->members; cnt; )
	if(ll = (LABELLIST *)LabelList->get(--cnt))
	{
		dbgout("#" << cnt << "\t" << ll->name << "\t" << (__int32)ll->sel_class << "\n")
		if(ll->name) free(ll->name);
	}
	dbgout("\n")

	delete LabelList;
}
